# -*- coding: utf-8 -*-
"""
   Wiki+

   Copyright (C) 2005 Yongjin Cho <yongjin.cho@gmail.com>
   Wiki+ is distributed under GNU GPL. See COPYING for details.

   This program is based on PikiPiki which is a simple wiki engine.
   PikiPiki project homepage is http://sf.net/projects/piki.

   For more information, visit http://wikiplus.kldp.net.
"""
# Imports -----------------------------------------------------------
import cgi, sys, string, os, fcntl, re, errno, time, urllib, Cookie
import config



# Exceptions --------------------------------------------------------

class WikiError(Exception):
    def __init__(self):
        self.title = ''
        self.contents = ''

    def send_error(self):
        send_title(self.title)
        print '<p>'
        print self.contents
        print '</p>'
        send_footer()


class NoPage(WikiError):               # 문서 자체가 없음
    def __init__(self, page_name):
        self.title = '없는 문서입니다.'
        self.contents = \
            '요청 하신 <b>%s</b>는 없는 문서 입니다. ' \
            '이 문서를 작성 하시려면 %s를 이용해 주세요.' % \
            (page_name,
             link_tag(urlencode(page_name) + '?action=edit', '새로 만들기'))


class DeletedPage(WikiError):         # 문서의 현재의 개정판이 없음
    def __init__(self, page_name):
        self.title = '삭제된 문서입니다.'
        self.contents = \
            '요청하신 %s는 삭제된 문서입니다. ' \
            '그렇지만 아직 예전 개정판이 남아있습니다. ' \
            '예전 개정판 정보를 보시려면 %s를 이용해 주시고, ' \
            '새로 작성하시려면 %s를 이용해 주세요.' % \
            (page_name,
             link_tag(urlencode(page_name) + '?action=history', '개정정보'),
             link_tag(urlencode(page_name) + '?action=edit', '문서 작성'))


class NoRevision(WikiError):           # 페이지의 특정 개정판이 없음
    def __init__(self, page_name, rtime):
        self.title = '없는 개정판입니다.'
        self.contents = \
            '%s 문서에 대한 %s의 개정판은 존재하지 않습니다. ' \
            '최근 개정판을 보시려면 %s를 이용해 주시고, ' \
            '다른 개정판을 보시려면 %s를 이용해 주세요.' % \
            (page_name, str_datetime(rtime),
             link_tag(urlencode(page_name), '보기'),
             link_tag(urlencode(page_name) + '?action=history', '개정정보 보기'))


class EmptyText(WikiError):
    def __init__(self):
        self.title = '내용이 없습니다.',
        self.contents = '내용이 없는 문서는 저장할 수 없습니다.'


class UnchangedText(WikiError):
    def __init__(self):
        self.title = '내용이 같습니다.'
        self.contents = '같은 내용의 문서는 저장할 수 없습니다.'


class UpdateConflict(WikiError):
    def __init__(self, page_name, newtext):
        self.title = '문서 수정 충돌!!!'
        self.contents = \
            '<b>아래 내용을 주의 깊게 읽어주십시요.</b><br />' \
            '이 페이지를 수정하는 동안 다른 사용자가 수정해서 ' \
            '먼저 저장했습니다. 작업하시던 내용은 아래 ' \
            '텍스트 박스에 그대로 있습니다. 이 내용을 지금 ' \
            '바로 덮어쓰신다면 다른 사용자가 먼저 작업한 내용을 ' \
            '덮어쓰게 됩니다. 따라서 이 내용을 메모장 같은 곳에 ' \
            '일단 복사하신 후 수정된 부분만을 추가해주시면 ' \
            '감사하겠습니다.<br /><br /><textarea wrap="virtual" ' \
            'rows="17" cols="80">%s</textarea><br /><br />' \
            '위의 내용을 메모장에 복사해 두셨다면 %s를 눌러서 ' \
            '다른 사용자가 수정한 것을 먼저 보시기 바랍니다.' % \
            (newtext, link_tag(urlencode(page_name), '여기'))


class DeniedIP(WikiError):
    def __init__(self):
        self.title = '금지된 IP 주소 입니다.'
        self.contents = \
            '해당 IP 주소는 관리자에 의해서 위키를 변경하는 작' + \
            '업을 하지 못하도록 막혔습니다.'


class UnknownAction(WikiError):
    def __init__(self, action):
        self.title = '알 수 없는 명령입니다.'
        self.contents = 'action의 전달 값 %s을 확인해주세요.' % action


# for internal use in Page class
NoText = 'NoText'



# Page class --------------------------------------------------------

class Page:
    """Page management class

       Directory Structure
       -------------------
        data/ --+-- text/ ---+-- pageid1
                |            `-- pageid2
                |
                +-- backup/ -+-- pageid1.rtime
                |            +-- pageid1.rtime
                |            `-- pageid2.rtime
                |
                `-- editlog

       Page ID
       -------
        quoted page name

       Editlog File Structure
       ----------------------
        +-------------------------------------+
        |page id<TAB>rtime<TAB>ip<NEWLINE>|
        |    .           .          .         |
        |                                     |
    """

    def __init__(self, name):
        self.name = name
        self.url_name = urlencode(name)
        self.file_name = self.url_name.replace('%', '_')

    def link(self):
        if self._has_current():
            return link_tag(self.url_name, self.name)
        else:
            return link_tag(self.url_name, '?') + self.name

    def update(self, newtext, rtime):
        # 사용자로 부터 넘겨받은 데이터를 정리
        ip = os.environ.get('REMOTE_ADDR', '')
        if ip in config.denied_ips:
            raise DeniedIP
        newtext = newtext.replace('\r', '')
        if not newtext:
            raise EmptyText

        # 현재 텍스트가 있는 경우의 처리
        if self._has_current():
            if rtime != self._current_rtime():
                raise UpdateConflict(self.name, newtext)
            elif newtext == self._text():
                raise UnchangedText
            os.rename(self._text_path(), os.path.join(config.backup_dir,
                                                self.file_name + '.' + rtime))

        # 새 텍스트를 기록
        text_file = open(self._text_path(), 'wb')
        text_file.write(newtext)
        text_file.close()
        os.chmod(self._text_path(), 0666)

        # 로그 파일에 추가
        rtime = self._current_rtime()
        editlog_add(self.file_name, rtime, ip)

    def send_page(self, rtime=''):
        # text를 읽어옴
        try:
            text = self._text(rtime)
        except NoText:
            if not rtime:
                if self._has_backup():
                    raise DeletedPage(self.name)
                else:
                    raise NoPage(self.name)
            else:
                if self._exists():
                    raise NoRevision(self.name, rtime)
                else:
                    raise NoPage(self.name)

        # title등 출력에 필요한 각종 정보를 생성
        title = re.sub(r'([a-z])([A-Z])', r'\1 \2', self.name)
        if not rtime:
            rtime = self._current_rtime()
            avail_act = ('reload', 'edit', 'history')
        else:
            title = '"%s"의 개정판' % title
            avail_act = ('history',)

        # 문서 출력
        send_title(title)
        send_body(self.name, text)
        send_footer(self.name, rtime, avail_act)

    def send_editor(self, preview=False, newtext='', rtime=''):
        if not preview:
            if self._exists():
                title = '"%s" 고치기' % self.name
                rtime = self._current_rtime()
                newtext = self._text()
            else:
                title = '"%s" 새로 만들기' % self.name
                rtime = ''
                newtext = ''
        else:
            title = '"%s" 미리보기' % self.name

        send_title(title)
        print '<form method="post" action="%s/%s#preview"><p>' % \
              (get_scriptname(), self.url_name)
        print '<input type="hidden" name="action" value="update" />'
        print '<input type="hidden" name="rtime" value="%s" />' % rtime
        print '<textarea name="newtext" rows="18" cols="90" ' \
              'style="width:95%%;word-wrap:virtual;">%s</textarea>' % newtext
        print '<br />필명: <input type="text" name="name" /> ' 
        print '<input type="submit" name="save" value="저장하기" /> ' \
              '<input type="submit" name="preview" value="미리보기" />'
        print link_tag(self.url_name, '취소하기')
        print '</p></form>'
        if preview:
            print '<a name="preview"></a><hr />'
            send_body(self.name, newtext)
        send_footer(self.name)

    def send_history(self):
        def print_row(rtime):
            if not rtime:
                rtime = self._current_rtime()
                size = self._text_size()
                suffix = ''
            else:
                suffix = '&rtime=' + rtime
                size = self._text_size(rtime)
            print '<tr><td>%s</td><td>%s</td><td>' % \
                  (str_datetime(rtime), str_size(size))
            print link_tag(self.url_name + '?action=show' + suffix, '보기')
            print link_tag(self.url_name + '?action=text' + suffix, '텍스트')
            print '</td></tr>'

        blist = self._backup_list()
        if not blist and not self._has_current():
            raise NoPage(self.name)
        blist.sort()
        blist.reverse()

        send_title('"%s"의 개정판 정보' % self.name)
        print '<table border="1" cellspacing="0" cellpadding="3"><tr>' \
              '<th>날짜</th><th>크기</th><th>명령</th></tr>'
        if self._has_current():
            print_row('')
        for rtime in blist:
            print_row(rtime)
        print '</table>'
        send_footer(self.name)

    def send_diff(self, rtime1, rtime2=''):
        pass
        
    def send_text(self, rtime=''):
        try:
            text = self._text(rtime)
        except NoText:
            if not rtime:
                if self._has_backup():
                    raise DeletedPage(self.name)
                else:
                    raise NoPage(self.name)
            else:
                if self._exists():
                    raise NoRevision(self.name, rtime)
                else:
                    raise NoPage(self.name)

        print '<pre>'
        print text
        print '</pre>'

    def _exists(self):
        return self._has_current() or self._has_backup()

    def _has_current(self):
        return os.path.exists(self._text_path())

    def _has_backup(self):
        return len(self._backup_list()) > 0

    def _text_path(self, rtime=''):
        if not rtime:
            return os.path.join(config.text_dir, self.file_name)
        else:
            return os.path.join(config.backup_dir,
                                self.file_name + '.' + rtime)

    def _text_size(self, rtime=''):
        try:
            return os.path.getsize(self._text_path(rtime))
        except OSError, er:
            if er.errno == errno.ENOENT:
                raise NoText
            else:
                raise er
    def _text(self, rtime=''):
        try:
            text_file = open(self._text_path(rtime), 'rb')
        except IOError, er:
            if er.errno == errno.ENOENT:
                raise NoText
            else:
                raise er
        try:
            text = text_file.read()
        finally:
            text_file.close()
        return text
            
    def _backup_list(self):
        tmp_list = os.listdir(config.backup_dir)
        tmp_list = filter(lambda x: x.startswith(self.file_name + '.'),
                          tmp_list)
        return map(lambda x: x.split('.')[1], tmp_list)

    def _current_rtime(self):
        try:
            return str(os.path.getmtime(self._text_path()))
        except OSError, er:
            if er.errno == errno.ENOENT:
                raise NoText
            else:
                raise er
        


def get_page_list():
    tmp_list = map(lambda x:x.replace('_', '%'), os.listdir(config.text_dir))
    return map(urldecode, tmp_list)



# User --------------------------------------------------------------

# The class to manage user infomation, login and logout

class User:
    def __init__(self, name):
        self.name = name

    def login(self, password):
        pass

    def logout(self):
        pass


# Editlog -----------------------------------------------------------

# Functions to keep track of when people have changed pages, so we can
# do the recent changes page and so on.
# The editlog is stored with one record per line, as tab-separated
# words: pageid, rtime, host, comment

def editlog_add(pageid, rtime, host):
    entry = string.join((pageid, rtime, host), "\t") + "\n"
    editlog = open(config.editlog_path, 'ab')
    try: 
        fcntl.flock(editlog.fileno(), fcntl.LOCK_EX)
        editlog.seek(0, 2)                  # to end
        editlog.write(entry)
    finally:
        fcntl.flock(editlog.fileno(), fcntl.LOCK_UN)
        editlog.close()


def editlog_raw_lines():
    editlog = open(config.editlog_path, 'rb')
    try:
        fcntl.flock(editlog.fileno(), fcntl.LOCK_SH)
        return editlog.readlines()
    finally:
        fcntl.flock(editlog.fileno(), fcntl.LOCK_UN)
        editlog.close()



# Formatter ---------------------------------------------------------

class Formatter:
    BLOCK_PAT = r'^(?:' \
        r'(?P<h2>=\s.*\s=)|' \
        r'(?P<h3>==\s.*\s==)|' \
        r'(?P<h4>===\s.*\s===)|' \
        r'(?P<h5>====\s.*\s====)|' \
        r'(?P<h6>=====\s.*\s=====)|' \
        r'(?P<hr>-{4,})|' \
        r'(?P<ul>\s+\*\s.*)|' \
        r'(?P<ol>\s+\d+\.\s.*)|' \
        r'(?P<dl>\s+[^(?:::)]*::.*)|' \
        r'(?P<table>\|\|(?:.*\|\|)+)|' \
        r'(?P<blockquote>"".*)|' \
        r'(?P<blank>\s*)|' \
        r'(?P<p>.+)' \
        r')$'

    URL_PAT = r'(?:http|https|ftp|news|mailto)\:[^\s\'"]+\S'
    EMAIL_PAT = r'[-\w._+]+\@[\w.-]+'
    INLINE_PAT = \
        r"(?P<emph>'{2,3})|" \
        r'(?P<word>\b(?:[A-Z][a-z]+){2,}\b)|' \
        r'(?P<eword>\[[^\[\]]+\])|' \
        r'(?P<url>' + URL_PAT + r')|' \
        r'(?P<email>' + EMAIL_PAT + r')|' \
        r'(?P<macro>\[\[[A-Za-z][A-Za-z0-9]*(?:\([^\)]*\))?\]\])|' \
        r'(?P<tag><i>|</i>|<b>|</b>|<tt>|</tt>|<big>|</big>|<small>|' \
        r'</small>|<sub>|</sub>|<sup>|</sup>|<ins>|</ins>|<del>|</del>)|' \
        r'(?P<br><br\s*/?>)|' \
        r'(?P<code>\{\{\{[^(?:\}\}\})]*\}\}\})|' \
        r'(?P<pre>\{\{\{|\}\}\})|' \
        r'(?P<ent>[<>&])'

    def __init__(self, page_name):
        self.page_name = page_name

        self._block_re = re.compile(Formatter.BLOCK_PAT)
        self._inline_re = re.compile(Formatter.INLINE_PAT)
        self._url_re = re.compile(Formatter.URL_PAT)
        self._email_re = re.compile(Formatter.EMAIL_PAT)
        self._indent_re = re.compile(r'^\s+')
        self._stack = [(None, None)] # (tag, data)

    def print_xhtml(self, text):
        lines = text.split('\n')
        for line in lines:
            if not self._in('pre'):
                self._block(line)
            else:
                self._inline(line)
        self._close_block('blank')

    def _close_all(self):
        self._close_inline()
        self._close_block('blank')

    # stack operations --------
    def _in(self, tag):
        if self._stack[-1][0] == tag:
            return True
        else:
            return False

    def _push(self, tag, data=None):
        self._stack.append((tag, data))

    def _pop(self):
        return self._stack.pop()

    def _pop_and_push(self, tag, data=None):
        tmp = self._stack.pop()
        self._stack.append((tag, data))
        return tmp

    def _peek(self):
        return self._stack[-1]

    def _peek_tag(self):
        return self._stack[-1][0]

    # block-level --------
    def _block(self, s):
        m = self._block_re.search(s)
        tag = m.lastgroup
        self._close_block(tag)
        apply(getattr(self, '_' + tag), (m.group(),))

    def _close_block(self, tag):
        list_tags = ('ul', 'ol', 'dl')

        upper_tag = self._peek_tag()
        if upper_tag in list_tags and tag not in list_tags:
            while upper_tag in list_tags:
                upper_tag, upper_level = self._pop()
                if upper_tag == 'dl':
                    print '</dd></dl>'
                else:
                    print '</li></%s>' % upper_tag
                upper_tag = self._peek_tag()

        elif upper_tag == 'table' and tag != 'table':
            self._pop()
            print '</tr></table>'

        elif upper_tag == 'p' and tag != 'p':
            self._pop()
            print '</p>'

        elif upper_tag == 'blockquote' and tag != 'blockquote':
            self._pop()
            print '</p></blockquote>'

    def _h2(self, s):
        self._h(s, 2)
    def _h3(self, s):
        self._h(s, 3)
    def _h4(self, s):
        self._h(s, 4)
    def _h5(self, s):
        self._h(s, 5)
    def _h6(self, s):
        self._h(s, 6)
    def _h(self, s, level):
        print '<h%d>' % level
        self._inline(s[level:-level])
        print '</h%d>' % level
        
    def _hr(self, s):
        print '<hr />'

    def _ul(self, s):
        self._l(s, 'ul')
    def _ol(self, s):
        self._l(s, 'ol')
    def _dl(self, s):
        self._l(s, 'dl')
    def _l(self, s, tag):
        indent = self._indent_re.match(s).end()
        upper_tag, upper_indent = self._peek()
        if upper_tag not in ('ul', 'ol', 'dl'):
            upper_indent = 0

        while indent < upper_indent:
            upper_tag, upper_indent = self._pop()
            if upper_tag == 'dl':
                print '</dd></%s>' % upper_tag
            else:
                print '</li></%s>' % upper_tag
            upper_tag, upper_indent = self._peek()
            if upper_tag not in ('ul', 'ol', 'dl'):
                upper_indent = 0
        if indent == upper_indent:
            if upper_tag == 'dl':
                print '</dd>'
                if tag in ('ul', 'ol'):
                    print '</dl><%s>' % tag
                    self._pop_and_push(tag, indent)
            else:
                print '</li>'
                if tag == 'dl':
                    print '</%s><dl>' % upper_tag
                    self._pop_and_push(tag, indent)
        elif indent > upper_indent:
            print '<%s>' % tag
            self._push(tag, indent)

        if tag == 'ul':
            print '<li>'
            self._inline(s[indent + 2:])
        elif tag == 'ol':
            pos = s.find('.')
            print '<li>'
            self._inline(s[pos + 2:])
        else:
            start = self._indent_re.match(s).end()
            end = s.find('::')
            print '<dt>'
            self._inline(s[start:end])
            print '</dt>'
            print '<dd>'
            self._inline(s[end + 2:])

    def _table(self, s):
        upper_tag, upper_ncols = self._peek()
        ncols = s.count('||')
        if upper_tag != 'table':
            print '<table border="1" cellspacing="0" cellpadding="3"><tr>'
            self._push('table', ncols)
        else:
            if ncols == upper_ncols:
                print '</tr><tr>'
            else:
                print '</tr></table>' \
                      '<table border="1" cellspacing="0" cellpadding="3"><tr>'
                self._push('table', ncols)

        cols = s[2:-2].split('||')
        for col in cols:
            print '<td>'
            self._inline(col)
            print '</td>'

    def _blockquote(self, s):
        upper_tag = self._peek_tag()
        if upper_tag != 'blockquote':
            print '<blockquote><p>'
            self._push('blockquote')
        self._inline(s[2:])
        
    def _blank(self, s):
        pass

    def _p(self, s):
        upper_tag = self._peek_tag()
        if upper_tag != 'p':
            print '<p>'
            self._push('p')
        self._inline(s)

    # inline --------
    def _inline(self, s):
        print self._inline_re.sub(self._replace, s)
        self._close_inline()

    def _close_inline(self):
        tags = ('i', 'b', 'tt', 'big', 'small', 'sub', 'sup', 'ins', 'del')
        upper_tag = self._peek_tag()
        while upper_tag in tags:
            self._pop()
            print '</%s>' % upper_tag            
            upper_tag = self._peek_tag()

    def _replace(self, m):
        return apply(getattr(self, '_' + m.lastgroup + '_repl'), (m.group(),))

    def _emph_repl(self, s):
        if not self._in('pre'):
            if len(s) == 3:
                if self._in('b'):
                    self._pop()
                    return '</b>'
                else:
                    self._push('b')
                    return '<b>'
            else:
                if self._in('i'):
                    self._pop()
                    return '</i>'
                else:
                    self._push('i')
                    return '<i>'
        else:
            return s

    def _word_repl(self, s):
        if not self._in('pre'):
            return Page(s).link()
        else:
            return s

    def _eword_repl(self, s):
        if not self._in('pre'):
            s = s[1:-1]
            m = self._url_re.match(s)
            if m:
                url = m.group()
                name = s[m.end():].strip()
                if not name:
                    name = url
                ext = url[url.rfind('.') + 1:]
                if ext in ('gif', 'jpg', 'bmp', 'png'):
                    return '<img src="%s" alt="%s" />' % (url, name)
                else:
                    return '<a href="%s">%s</a>' % (url, name)

            m = self._email_re.match(s)
            if m:
                email = m.group()
                name = s[m.end():].strip()
                if not name:
                    name = email
                return '<a href="mailto:%s">%s</a>' % (email, name)

            return Page(s).link()
        else:
            return s

    def _url_repl(self, s):
        if not self._in('pre'):
            ext = s[s.rfind('.') + 1:]
            if ext in ('gif', 'jpg', 'bmp', 'png'):
                return '<img src="%s" alt="%s" />' % (s, s)
            else:
                return '<a href="%s">%s</a>' % (s, s)
        else:
            return s

    def _email_repl(self, s):
        if not self._in('pre'):
            return '<a href="mailto:%s">%s</a>' % (s, s)
        else:
            return s

    def _macro_repl(self, s):
        if not self._in('pre'):
            pos = s.find('(')
            if pos == -1:
                name = s[2:-2]
                arg = ''
            else:
                name = s[2:pos]
                arg = s[pos+1:-3]
            try:
                method = getattr(self, '_macro_' + name)
            except:
                return '<i>%s 매크로는 없습니다.</i>' % name
            return method(arg)
        else:
            return s

    def _code_repl(self, s):
        if not self._in('pre'):
            s = s.replace('&', '&amp;').replace('<', '&lt;')
            s = s.replace('>', '&gt;')
            return '<tt>%s</tt>' % s[3:-3]
        else:
            return s

    def _pre_repl(self, s):
        if not self._in('pre') and s == '{{{':
            self._close_inline()
            self._close_block('blank')  # pre에 들어가기전 block을 닫아줘야한다
            self._push('pre')
            return '<pre>'
        elif self._in('pre') and s == '}}}':
            self._pop()
            return '</pre>'
        else:
            return s

    def _ent_repl(self, s):
        return {'&': '&amp;', '<': '&lt;', '>': '&gt;'}[s]


    def _br_repl(self, s):
        if not self._in('pre'):
            return '<br />'
        else:
            return '&lt;br /&gt;'

    def _tag_repl(self, s):
        is_open = (True, False)[s[1] == '/']
        if is_open:
            self._push(s[1:-1])
            return s
        else:
            upper_tag = self._peek_tag()
            tag = s[2:-1]
            if tag == upper_tag:
                self._pop()
                return s
            else:
                return ''
        
    # macros --------
    def _macro_Echo(self, arg):
        arg = arg.replace('&', '&amp;').replace('<', '&lt;')
        arg = arg.replace('>', '&gt;')
        return arg
        
        
    def _macro_RecentChanges(self, arg):
        import cStringIO

        self._close_all()
        lines = editlog_raw_lines()
        lines.reverse()
    
        buf = cStringIO.StringIO()
        ratchet_day = None
        done_names = {}
        for line in lines:
            page_name, rtime, ip = line[:-1].split('\t')
            page_name = unquote_name(page_name)

            if done_names.has_key(page_name):
                continue
            done_names[page_name] = 1

            # year, month, day, DoW
            time_tuple = time.localtime(float(rtime))
            day = tuple(time_tuple[0:3])
            if day != ratchet_day:
                if ratchet_day:
                    buf.write('</dl>')
                buf.write('<dl><dt><b>%s</b></dt>' % time.strftime(config.date_fmt, time_tuple))
                ratchet_day = day

            buf.write('<dd>')
            buf.write(Page(page_name).link())
            buf.write(' . . . . ')
            if config.show_ips:
                buf.write(ip + ' ')

            if config.changed_time_fmt:
                buf.write(time.strftime(config.changed_time_fmt, time_tuple))

            buf.write('</dd>')
        buf.write('</dl>')

        return buf.getvalue()

    
    def _macro_FullSearch(self, arg):
        self._close_all()
        url_name = urlencode(self.page_name)
        return '<form action="%s/%s" method="get">' \
               '<input type="hidden" name="action" value="fullsearch" />' \
               '<input type="text" name="value" size="30" /> ' \
               '<input type="submit" value="Search" />' \
               '</form>' % (get_scriptname(), url_name)


    def _macro_TitleSearch(self, arg):
        self._close_all()
        url_name = urlencode(self.page_name)
        return '<form action="%s/%s" method="get">' \
               '<input type="hidden" name="action" value="titlesearch" />' \
               '<input type="text" name="value" size="30" /> ' \
               '<input type="submit" value="Search" />' \
               '</form>' % (get_scriptname(), url_name)


    def _macro_TitleIndex(self, arg):
        import cStringIO

        self._close_all()

        page_list = get_page_list()
        tmp_list = map(lambda x: x.decode(config.encoding), page_list)
        tmp_list = map(lambda x: (x.upper(), x), tmp_list)
        tmp_list.sort()
        page_list = map(lambda x: x[1], tmp_list)

        buf = cStringIO.StringIO()
        index_letters = []
        current_letter = None
        for name in page_list:
            letter = get_index_group(name)
            letter = letter.upper()
            if letter not in index_letters:
                index_letters.append(letter)

            if letter != current_letter:
                if current_letter:
                    buf.write('</dl>')
                encoded_letter = letter.encode(config.encoding)
                buf.write('<dl><dt><b><a name="%s" href="#top">%s</a></b></dt>' %
                                                (encoded_letter, encoded_letter))
                current_letter = letter

            buf.write('<dd>%s</dd>' % Page(name.encode(config.encoding)).link())
        buf.write('</dl>')

        index = map(lambda x: x.encode(config.encoding), index_letters)
        index = map(lambda x: '<a href="#%s">%s</a> ' % (x, x), index)
        top = '<a name="top"></a>'

        return '<p>' + top + ' | '.join(index) + '</p>' + buf.getvalue()


    def _macro_PageCount(self, arg):
        page_list = get_page_list()
        return str(len(page_list))


    def _macro_DueDate(self, arg):
        import math
        try:
            t = time.strptime(arg, '%Y%m%d')
        except ValueError, er:
            return '<i>%s</i>' % er

        secs = time.mktime(t)
        cur_secs = time.time()
        dist_day = int(math.ceil((secs - cur_secs) / (24 * 60 * 60)))

        return str(abs(dist_day))


    def _macro_UploadFile(self, arg):
        self._close_all()
        return '<form action="%s/%s" method="post" enctype="multipart/form-data">' \
               '<input type="hidden" name="action" value="upload" />' \
               '<input type="file" name="file" size="40" /><br />' \
               '파일이름(옵션): <input type="text" name="rename" size="30" />' \
               '<br /><input type="submit" value="업로드" />' \
               '</form>' % (get_scriptname(), self.page_name)


    def _macro_UploadedFile(self, arg):
        file_name = arg
        img_exts = ('.jpg', '.gif', '.bmp', '.png', 'jpeg')
        name = cgi.escape(file_name)

        if os.path.exists(os.path.join(config.upload_dir, file_name)):
            url = config.upload_url + '/' + urlencode(name)

            if os.path.splitext(file_name)[1] in img_exts:
                return '<img src="%s" alt="%s" />' % (url, name)
            else:
                return '<a href="%s">%s</a>' % (url, name)
        else:
            return '<i>"%s" 파일이 없습니다.</i>' % name


    def _macro_UploadedFiles(self, arg):
        import cStringIO

        self._close_all()
        buf = cStringIO.StringIO()

        file_list = os.listdir(config.upload_dir)
        buf.write('<table border="1" cellspacing="0" cellpadding="3"><tr>')
        buf.write('<th>파일 이름</th><th>크기</th><th>날짜</th></tr>')
        for file_name in file_list:
            file_path = os.path.join(config.upload_dir, file_name)

            name = cgi.escape(file_name)
            url = config.upload_url + '/' + urlencode(file_name)
            size = os.path.getsize(file_path)
            mtime = os.path.getmtime(file_path)

            buf.write('<tr><td><a href="%s">%s</a></td>' % (url, name))
            buf.write('<td align="right">%s</td>' % str_size(size))
            buf.write('<td>%s</td></tr>' % \
                      time.strftime(config.datetime_fmt, time.localtime(mtime)))
        buf.write('</table>')
        buf.write('<p>총 %d개 파일이 업로드되었습니다.</p>' % len(file_list))

        return buf.getvalue()


    def _macro_ISBN(self, arg):
        isbn = arg
        amazon_link = 'http://www.amazon.com/exec/obidos/ISBN=%s'
        amazon_img = 'http://images.amazon.com/images/P/%s.01.MZZZZZZZ.gif'
        aladdin_link = 'http://www.aladdin.co.kr/catalog/book.asp?ISBN=%s'
        aladdin_img = 'http://www.aladdin.co.kr/Cover/%s_1.gif'

        if isbn[:2] == '89':
            link, img = aladdin_link % isbn, aladdin_img % isbn
        else:
            link, img = amazon_link % isbn, amazon_img % isbn
    
        return '<a href="%s"><img src="%s" alt="%s" /></a>' % (link, img, isbn) 




# Utility functions -------------------------------------------------

def emit_header():
    print 'Content-type: text/html; charset=%s' % config.encoding
    print


def urlencode(s):
    t = urllib.quote(s, '')
    t = t.replace('-', '%2D')
    t = t.replace('.', '%2E')
    return t.replace('_', '%5F')


def urldecode(s):
    return urllib.unquote(s)


def quote_name(name):
    return urlencode(name).replace('%', '_')


def unquote_name(name):
    return urldecode(name.replace('_', '%'))


def quote_path(path):
    path = path.replace('/', '_')
    path = path.replace('\\', '_')
    return path.replace(':', '_')


def get_scriptname():
    return os.environ.get('SCRIPT_NAME', '')


def get_index_group(name):
    c = name[0]
    if u'\uAC00' <= c <= u'\uD7AF': # Hangul Syllables
        return unichr(0xac00 + (int(ord(c) - 0xac00) / 588) * 588)
    else:
        return c


def get_basename(path):
    import posixpath, dospath, macpath
    for pathmodule in (posixpath, dospath, macpath):
        basename = pathmodule.basename(path)
        if basename != path:
            return basename
    return path


def str_size(size):
    if size < 1024:
        return str(size)
    elif size < 1024 ** 2:
        ksize = size / 1024.0
        if int(ksize) < 10:
            return str(round(ksize, 1)) + 'K'
        return str(int(ksize)) + 'K'
    else:
        msize = size / (1024.0 ** 2)
        if int(msize) < 10:
            return str(round(msize, 1)) + 'M'
        return str(int(msize)) + 'M'


def str_datetime(rtime):
    return time.strftime(config.datetime_fmt, time.localtime(int(rtime)))


def str_date(rtime):
    return time.strftime(config.date_fmt, time.localtime(int(rtime)))


def link_tag(link, text=None):
    if not text:
        text = link
    return '<a href="%s/%s">%s</a>' % (get_scriptname(), link, text)


_is_admin = None
def is_admin():
    global _is_admin

    if _is_admin == None:
        if os.environ.has_key('HTTP_COOKIE'):
            c = Cookie.SimpleCookie()
            c.load(os.environ['HTTP_COOKIE'])
            if c.has_key('key'):
                try:
                    f = open(config.session_path, 'r')
                except IOError:
                    _is_admin = False
                else:
                    key = f.read()
                    f.close()
                    if c['key'].value == key:
                        _is_admin = True
                    else:
                        _is_admin = False
            else:
                _is_admin = False
        else:
            _is_admin = False

    return _is_admin


def send_title(title):
    print '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" ' \
          '"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd"><html>'
    print '<head><title>%s</title>' % ' - '.join((title, config.wiki_name))
    print '<meta http-equiv="content-type" content="text/html;' \
          'charset=%s" />' % config.encoding
    print '<link rel="stylesheet" type="text/css" href="%s" /></head>' \
          '<body>' % config.css_url
    print '<div class="menu">%s</div>' % config.menu_bar
    print '<div>'
    print link_tag(urlencode(config.wiki_name),
               '<img class="logo" src="%s" alt="[HOME]" />' % config.logo_url)
    print '</div><h1>%s</h1><div><br /></div>' % title


def send_body(page_name, text):
    Formatter(page_name).print_xhtml(text)

def send_footer(page_name='', rtime=None, avail_act=None):
    print '<hr />'
    print '<div class="wikiplus">Powered by ' \
          '<a href="http://wikiplus.kldp.net">Wiki+</a><br />' \
          '<a href="http://validator.w3.org/check/referer">XHTML 1.0</a> | ' \
          '<a href="http://jigsaw.w3.org/css-validator/check/referer">' \
          'CSS 2.0</a></div><div>'
    if rtime:
        print '이 문서는 %s에 수정되었습니다.<br />' % \
              str_datetime(rtime)
    if avail_act:
#        print '이 문서에 사용 가능한 명령은 다음과 같습니다.<br />'
        url_name = urlencode(page_name)
        if 'reload' in avail_act:
            print link_tag(url_name, '새로 고침')
        if 'show' in avail_act:
            print link_tag(url_name, '보기')
        if 'edit' in avail_act:
            print link_tag(url_name + '?action=edit', '고치기')
        if 'history' in avail_act:
            print link_tag(url_name + '?action=history', '개정정보 보기')
        if is_admin():
            print link_tag(url_name + '?action=logout', '관리자 로그아웃')
        else:
            print link_tag(url_name + '?action=loginform', '관리자 로그인')
    print '</div></body></html>'


def send_refresh(page_name):
    emit_header()
    print """<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html>
<head><meta http-equiv="refresh" content="0;url=%s/%s"></head>
<body></body>
</html>""" % (get_scriptname(), urlencode(page_name))



# Action handlers ---------------------------------------------------

def do_show(page_name, form):
    emit_header()

    pg = Page(page_name)
    try:
        pg.send_page(form.getvalue('rtime', ''))
    except (NoPage, DeletedPage, NoRevision), e:
        e.send_error()


def do_edit(page_name, form):
    emit_header()

    Page(page_name).send_editor()


def do_update(page_name, form):
    emit_header()

    pg = Page(page_name)
    newtext = form.getvalue('newtext', '').replace('\r', '')
    rtime = form.getvalue('rtime', '')
    if form.getvalue('save'):
        try:
            pg.update(newtext, rtime)
        except (EmptyText, UnchangedText, UpdateConflict, DeniedIP), e:
            e.send_error()
        else:
            pg.send_page()
    else:   # if preview
        pg.send_editor(True, newtext, rtime)


def do_history(page_name, form):
    emit_header()

    pg = Page(page_name)
    try:
        pg.send_history()
    except NoPage, e:
        e.send_error()
       

def do_diff(page_name, form):
    emit_header()

    pg = Page(page_name)
    try:
        pg.send_diff()
    except (NoPage, NoRevision), e:
        e.send_error()

        
def do_text(page_name, form):
    emit_header()

    pg = Page(page_name)
    try:
        pg.send_text(form.getvalue('rtime', ''))
    except (NoPage, DeletedPage, NoRevision), e:
        e.send_error()


def do_goto(page_name, form):
    emit_header()

    page_name = form.getvalue('value')
    init_macro(page_name, form)         # reset global variables for macro
    Page(page_name).send_page()


def do_fullsearch(page_name, form):
    emit_header()

    _do_search(page_name, form, 'full')

def do_titlesearch(page_name, form):
    emit_header()

    _do_search(page_name, form, 'title')

def _do_search(page_name, form, method):
    needle = form.getvalue('value', '')
    if method == 'title':
        send_title('"%s"로 제목만 검색한 결과' % needle)
    else:
        send_title('"%s"로 검색한 결과' % (needle))

    needle_re = re.compile(needle, re.IGNORECASE)
    page_list = get_page_list()
    title_hits = filter(needle_re.search, page_list)

    contents_hits = []
    if method != 'title':
        for page in page_list:
            body = Page(page)._text()
            match = needle_re.findall(body)
            count = len(needle_re.findall(body))
            if count:
                contents_hits.append((count, page))
        contents_hits.sort()
        contents_hits.reverse()

    print '<dl>'
    print '<dt><b>제목 검색 결과: %d건</b></dt>' % len(title_hits)
    for page in title_hits:
        print '<dd>' + Page(page).link() + '</dd>'
    print '</dl>'
    if method != 'title':
        print '<dl>'
        print '<dt><b>본문 검색 결과: %d건</b></dt>' % len(contents_hits)
        for (count, page) in contents_hits:
            print '<dd>' + Page(page).link()
            print ' . . . . %d번' % count
        print '</dl>'

    print '<p>총 %d의 문서를 검색하였습니다.</p>' % len(page_list)
    send_footer(page_name, 1)


def do_upload(page_name, form):
    import shutil

    emit_header()

    if not form.getvalue('file') or not form['file'].file:
        msg = '<b>File was not trasferred.</b>'
    else:
        base_name = form.getvalue('rename')
        if not base_name:
            base_name = get_basename(form['file'].filename)

        name, ext = os.path.splitext(quote_path(base_name))
        if ext in config.allowed_ext:
            # conflict resolution
            suffix = ''
            while os.path.exists(os.path.join(config.upload_dir,
                                              name + suffix + ext)):
                if not suffix:
                    suffix = '0'
                else:
                    suffix = str(int(suffix) + 1)
            file_name = name + suffix + ext            
            file_path = os.path.join(config.upload_dir, file_name)

            # copy
            dest_file = open(file_path, 'wb')
            shutil.copyfileobj(form['file'].file, dest_file, 1024)
            dest_file.close()
            os.chmod(file_path, 0666)

            msg = '<b>"%s" was successfully uploaded.</b>' % file_name
        else:
            msg = '<b>"%s" is not allowed extension.</b>' % ext
    
    Page(page_name).send_page()


def do_loginform(page_name, form):
    emit_header()

    url_name = urlencode(page_name)
    print '<form action="%s/%s" method="get">' \
          '<input type="hidden" name="action" value="login" />' \
          '관리자 암호: <input type="password" name="password" size="30" />' \
          '<input type="submit" value="로그인" />' \
          '</form>' % (get_scriptname(), url_name)


def do_login(page_name, form):
    if form['password'].value == config.password:
        import md5

        m = md5.new()
        m.update(`time.time()`)
        key = m.digest()

        f = open(config.session_path, 'w+')
        f.write(key)
        f.close()

        c = Cookie.SimpleCookie()
        c['key'] = key
        c['key']['domain'] = '.yongd.com'
        c['key']['path'] = '/'
        c['key']['expires'] = Cookie._getdate(3600)
        print c

    send_refresh(page_name)


def do_logout(page_name, form):
    f = open(config.session_path, 'w+')
    f.close()

    c = Cookie.SimpleCookie()
    c['key'] = ''
    c['key']['max-age'] = 0
    print c

    send_refresh(page_name)


def get_handler(action):
    handlers = {
        'show':         do_show,
        'edit':         do_edit,
        'update':       do_update,
        'history':      do_history,
        'diff':         do_diff,
        'text':         do_text,
        'goto':         do_goto,
        'fullsearch':   do_fullsearch,
        'titlesearch':  do_titlesearch,
        'upload':       do_upload,
        'loginform':    do_loginform,
        'login':        do_login,
        'logout':       do_logout
    }
    if not handlers.has_key(action):
        raise UnknownAction(action)
    return handlers[action]



# CGI main ----------------------------------------------------------

def main():
    try:
        # get request
        path_info = os.environ.get('PATH_INFO', '')
        if len(path_info) > 1:
            page_name = path_info[1:]
        else:
            page_name = config.wiki_name
        form = cgi.FieldStorage()
        action = form.getvalue('action', 'show')

        # handle request
        handler = get_handler(action)
        handler(page_name, form)

    except UnknownAction, e:
        e.send_error()

    except:
        cgi.print_exception()

    sys.stdout.flush()
